from dune.perftool.ufl.modified_terminals import Restriction

from dune.perftool.options import get_option

from dune.perftool.pdelab.argument import name_coefficientcontainer
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.generation import (class_member,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      iname,
                                      include_file,
                                      initializer_list,
                                      instruction,
                                      loopy_class_member,
                                      preamble,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      valuearg
                                      )
from dune.perftool.loopy.buffer import get_buffer_temporary
from dune.perftool.pdelab.localoperator import (name_domain_field,
                                                lop_template_range_field,
                                                )
from dune.perftool.pdelab.quadrature import quadrature_order
from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pytools import ImmutableRecord, product

import pymbolic.primitives as prim
import loopy as lp
import numpy


class AMatrix(ImmutableRecord):
    def __init__(self, rows, cols, transpose=False, derivative=False, face=None):
        ImmutableRecord.__init__(self,
                                 rows=rows,
                                 cols=cols,
                                 transpose=transpose,
                                 derivative=derivative,
                                 face=face,
                                 )

    @property
    def name(self):
        return name_theta(self.transpose, self.derivative, face=self.face)

    @property
    def vectorized(self):
        return False

    def output_to_pymbolic(self, name):
        return lp.TaggedVariable(name, "sumfac")


class LargeAMatrix(ImmutableRecord):
    def __init__(self, rows, cols, transpose, derivative, face):
        assert isinstance(derivative, tuple)
        ImmutableRecord.__init__(self,
                                 rows=rows,
                                 cols=cols,
                                 transpose=transpose,
                                 derivative=derivative,
                                 face=face,
                                 )

    @property
    def name(self):
        name = "ThetaLarge{}{}_{}_qp{}".format("face{}_".format(self.face) if self.face is not None else "",
                                               "T" if self.transpose else "",
                                               "x".join(tuple("d" if d else "" for d in self.derivative)),
                                               quadrature_points_per_direction(),
                                               )
        for i, d in enumerate(self.derivative):
            define_theta(name, (self.rows, self.cols), self.transpose, d, face=self.face, additional_indices=(i,))

        return loopy_class_member(name,
                                  classtag="operator",
                                  dtype=numpy.float64,
                                  dim_tags="f,f,vec",
                                  shape=(self.rows, self.cols, 4),
                                  potentially_vectorized=True,
                                  managed=True,
                                  )

    @property
    def vectorized(self):
        return True

    def output_to_pymbolic(self, name):
        return lp.TaggedVariable(name, "vector")


def quadrature_points_per_direction():
    # TODO use quadrature order from dune.perftool.pdelab.quadrature
    # as soon as we have a generic implementation

    # Quadrature order
    q = quadrature_order()

    # Quadrature points in per direction
    nb_qp = q // 2 + 1

    return nb_qp


def polynomial_degree():
    form = get_global_context_value("formdata").preprocessed_form
    return form.coefficients()[0].ufl_element()._degree


def basis_functions_per_direction():
    return polynomial_degree() + 1


def define_oned_quadrature_weights(name):
    loopy_class_member(name,
                       dtype=numpy.float64,
                       classtag="operator",
                       shape=(quadrature_points_per_direction(),),
                       )


def name_oned_quadrature_weights():
    num = quadrature_points_per_direction()
    name = "qw_num{}".format(num)
    define_oned_quadrature_weights(name)
    return name


def define_oned_quadrature_points(name):
    loopy_class_member(name,
                       dtype=numpy.float64,
                       classtag="operator",
                       shape=(quadrature_points_per_direction(),),
                       )


def name_oned_quadrature_points():
    num = quadrature_points_per_direction()
    name = "qp_num{}".format(num)
    define_oned_quadrature_points(name)
    return name


@class_member(classtag="operator")
def typedef_polynomials(name):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    degree = polynomial_degree()

    include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="operatorfile")

    # TODO: make switching polynomials possible
    # return "using {} = Dune::QkStuff::GaussLobattoLagrangePolynomials<{}, {}, {}>;".format(name,
    #                                                                                        domain_field,
    #                                                                                        range_field,
    #                                                                                        degree)
    return "using {} = Dune::QkStuff::EquidistantLagrangePolynomials<{}, {}, {}>;".format(name,
                                                                                          domain_field,
                                                                                          range_field,
                                                                                          degree)


def type_polynomials():
    name = "Polynomials1D"
    typedef_polynomials(name)
    return name


@class_member(classtag="operator")
def define_polynomials(name):
    polynomials_type = type_polynomials()
    return "{} {};".format(polynomials_type, name)


def name_polynomials():
    name = "poly"
    define_polynomials(name)
    return name


@preamble(kernel="operator")
def sort_quadrature_points_weights(qp, qw):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    number_qp = quadrature_points_per_direction()
    include_file("dune/perftool/sumfact/onedquadrature.hh", filetag="operatorfile")
    return "onedQuadraturePointsWeights<{}, {}, {}>({}, {});".format(range_field, domain_field, number_qp, qp, qw)


@iname(kernel="operator")
def theta_iname(name, bound):
    name = "{}_{}".format(name, bound)
    domain(name, bound, kernel="operator")
    return name


class PolynomialLookup(FunctionIdentifier):
    def __init__(self, pol, derivative):
        self.pol = pol
        self.derivative = derivative

    def __getinitargs__(self):
        return (self.pol, self.derivative)

    @property
    def name(self):
        return "{}.{}".format(self.pol, "dp" if self.derivative else "p")


@function_mangler
def polynomial_lookup_mangler(target, func, dtypes):
    if isinstance(func, PolynomialLookup):
        return CallMangleInfo(func.name, (NumpyType(numpy.float64),), (NumpyType(numpy.int32), NumpyType(numpy.float64)))


def define_theta(name, shape, transpose, derivative, face=None, additional_indices=()):
    qp = name_oned_quadrature_points()
    qw = name_oned_quadrature_weights()
    sort_quadrature_points_weights(qp, qw)
    polynomials = name_polynomials()

    dim_tags = "f,f"
    if additional_indices:
        dim_tags = dim_tags + ",c"
        shape = shape + (4,)

    loopy_class_member(name,
                       dtype=numpy.float64,
                       shape=shape,
                       classtag="operator",
                       dim_tags=dim_tags,
                       managed=True,
                       potentially_vectorized=True,
                       )

    i = theta_iname("i", shape[0])
    j = theta_iname("j", shape[1])

    inames = i, j
    if transpose:
        inames = j, i

    args = [prim.Variable(inames[1]), prim.Subscript(prim.Variable(qp), (prim.Variable(inames[0]),))]
    if face is not None:
        args[1] = face

    instruction(assignee=prim.Subscript(prim.Variable(name), (prim.Variable(i), prim.Variable(j)) + additional_indices),
                expression=prim.Call(PolynomialLookup(polynomials, derivative), tuple(args)),
                kernel="operator",
                )


def name_theta(transpose=False, derivative=False, face=None):
    name = "{}{}Theta{}_qp_{}".format("face{}_".format(face) if face is not None else "",
                                      "d" if derivative else "",
                                      "T" if transpose else "",
                                      quadrature_points_per_direction(),
                                      )

    shape = [quadrature_points_per_direction(), basis_functions_per_direction()]
    if face is not None:
        shape[0] = 1
    if transpose:
        shape = shape[1], shape[0]
    shape = tuple(shape)

    define_theta(name, shape, transpose, derivative, face=face)
    return name


def construct_amatrix_sequence(transpose=False, derivative=None, facedir=None, facemod=None):
    dim = world_dimension()
    result = [None] * dim

    for i in range(dim):
        rows = quadrature_points_per_direction()
        cols = basis_functions_per_direction()

        onface = None
        if facedir == i:
            rows = 1
            onface = facemod

        if transpose:
            rows, cols = cols, rows

        result[i] = AMatrix(rows, cols, transpose=transpose, derivative=derivative == i, face=onface)

    return tuple(result)
